/*
 * $Id: salvage.c,v 2.11 1999-09-26 01:06:36+09 yohta Exp $
 * salvage file fragment from raw disk device(stdin).
 *
 * Copyright (c) 1999 Yoshihiko -OHTA- SARUMARU <mistral@imasy.or.jp>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char copyright[] = 
"Copyright (c) 1999 Yoshihiko -OHTA- SARUMARU <mistral@imasy.or.jp>\n\
 * All rights reserved.\n";
#endif /* not lint */

#include <stdio.h>
#include <strings.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>

#define	MAX_MAGIC	64
#define ON		 1
#define OFF		 0
#define END		-1

struct {
    char *magic;
    int   offset;
    char *tailstr;
    int   tailsize;
    char *suffix;
    int	  flag;
} magics[MAX_MAGIC] = {
    { "\037\235",		0,	"\000",	1,	"Z",		ON,  },
    { "\037\213",		0,	"\000", 1,	"gz",		ON,  },
    { "-lh",			2,	"",	0,	"lzh",		ON,  },
    { "ustar",			257,	"",	0,	"tar",		OFF, },
    { "GIF8",                   0,	"",	0,	"gif",          ON,  },
    { "\xff\xd8",		0,	"",	0,	"jpg",		ON,  },
    { "#FIG",			0,	"",	0,	"fig",		ON,  },
    { "%PDF",			0,	"",	0,	"pdf",		ON,  },
    { "%!PS",			0,	"",	0,	"ps",		ON,  },
    { "<!DOCTYPE NETSCAPE-Bookmark-file-1>", 0,	"", 0,	"bookmark",	ON,  },
    { "<!DOCTYPE HTML",		0,	"",	0,	"html",		ON,  },
    { "<TABLE",			0,	"",	0,	"html3",	ON,  },
    { "<PRE>",			0,	"",	0,	"html2",	ON,  },
    { "\\document",		0,	"",	0,	"tex",		ON,  },
    { "\367\002",		0,	"",	0,	"dvi",		ON,  },
    { "#!/bin/sh",		0,	"",	0,	"sh",		ON,  },
    { "#!/bin/csh",		0,	"",	0,	"csh",		ON,  },
    { "#!/usr/bin/awk",		0,	"",	0,	"awk",		ON,  },
    { "#!/usr/bin/perl",	0,	"",	0,	"pl",		ON,  },
    { "#!/usr/local/bin/ruby",	0,	"",	0,	"rb",		ON,  },
    { "#!",			0,	"",	0,	"script",	ON,  },
    { "/*",			0,	"",	0,	"h",		ON,  },
    { "#include",		0,	"",	0,	"c",		ON,  },
    { "ҳ",			0,	"",	0,	"txt",		ON,  },
    { "From ",			0,	"",	0,	"mail",		ON,  },
    { "From:",			0,	"",	0,	"mh",		ON,  },
    { "\xd0\xcf",		0,	"",	0,	"xls",		OFF, },
    { "",			0,	"",	0,	"",		END, }
};


#define BACKSPACE	"\010"
#define BLOCKSIZE	512	

#define progress_baton(count) \
	write(STDERR_FILENO, BACKSPACE, strlen(BACKSPACE)); \
	if ((count) == 0) write(STDERR_FILENO, "|", strlen("|")); else \
	if ((count) == 1) write(STDERR_FILENO, "\\", strlen("\\")); else \
	if ((count) == 2) write(STDERR_FILENO, "-", strlen("-")); else \
	if ((count) == 3) write(STDERR_FILENO, "/", strlen("/"));

void
usage() {
    int c;
    fprintf(stderr, "Salvage file fragments from raw disk image(stdin).\n");
    fprintf(stderr, "    $Id: salvage.c,v 2.11 1999-09-26 01:06:36+09 yohta Exp $\n\n");
    fprintf(stderr, "Usage: salvage [(+|-)(gif|jpg|...|html] < disk image\n");
    fprintf(stderr, "    supported file types are:");
    for (c = 0; magics[c].flag != END; c++)
	fprintf(stderr, " (%s)%s", 
		(magics[c].flag) ? "+" : "-", magics[c].suffix);
    fprintf(stderr, ".\n\n");
    fprintf(stderr, "Example: salvage -jpg +mh < /dev/rwd1s1f\n");
    fprintf(stderr, "Any comments is welcome --> "
	            "Yoshihiko SARUMARU <mistral@imasy.or.jp>\n");
    exit(0);
}

void
ParseArgument(char *arg) {
    int c;
    for (c=0; magics[c].flag != END; c++) {
	if (strcmp(magics[c].suffix, &arg[1]) == 0) {
	    switch (arg[0]) {
	    case '-':
		magics[c].flag = OFF;
		break;
	    case '+':
		magics[c].flag = ON;
		break;
	    default:
		fprintf(stderr, "Unknown argument: %s\n", arg);
		break;
	    }
	    break;
	}
    }
    if (strcmp("all", &arg[1]) == 0) {
	int c;
	switch (arg[0]) {
	case '-':
	    for (c=0; magics[c].flag != END; c++)
		magics[c].flag = OFF;
	    break;
	case '+':
	    for (c=0; magics[c].flag != END; c++)
		magics[c].flag = ON;
	    break;
	default:
	    fprintf(stderr, "Unknown switch: %s\n", arg);
	    break;
	}
	return;
    } 
    if (strncmp(arg, "-v", 2) == 0 || strncmp(arg, "-h", 2) == 0) {
	usage();
    }
    if (magics[c].flag == END) {
	fprintf(stderr, "Unknown argument: %s\n", arg);
    }
}

int
main(int argc, char **argv) {
    FILE *fp = NULL;
    int files = 0;
    int count = 0;
    int size;
    int current_type;
    char filename[FILENAME_MAX];
    char buf[BLOCKSIZE];
    int argc_org = argc; 

    while (argc-- > 1) {
	ParseArgument(argv[argc_org - argc]);
    }

    fprintf(stderr, "Salvage for:");
    for (count = 0; magics[count].flag != END; count++)
	if (magics[count].flag == ON) 
	    fprintf(stderr, " %s", magics[count].suffix);
    fprintf(stderr, ".\n");

    /* set priority to very low (= high nice value) */
    setpriority(PRIO_PROCESS, 0, 20);
    fprintf(stderr, "current priority is %d\n", getpriority(PRIO_PROCESS, 0));

    while( ! feof(stdin) ) {
	int type;
	
	memset(buf, 0, sizeof(buf));
	fread(buf, 1, BLOCKSIZE, stdin);
	if ((count ++ % 300) == 0) { progress_baton((count/300) % 4); };

	type = 0;
	while (magics[type].flag != END) {
	    if ((magics[type].flag == OFF) ||
		(strncmp(&buf[0+magics[type].offset], magics[type].magic, 
			strlen(magics[type].magic)) != 0) ) {
		type ++;
	    } else {
		if (fp != NULL) {
		    /* detect next file start before previous file end */
		    fclose(fp);
		    fprintf(stderr, "%s %6d bytes(%4.1f blocks) salvaged"
			    "(truncated or verbose).\n", BACKSPACE,
			    size, (double)size/BLOCKSIZE);
		}
		sprintf(filename, "%d.%s", files, magics[type].suffix);
		fp = fopen(filename, "w");
		fprintf(stderr, "%s%3d.%-5s: ", BACKSPACE, files, 
			magics[type].suffix);
		size = 0;
		files ++;
		current_type = type;
		break;
	    }
	}
	if (fp != NULL) {
		/* block (not) ends with 8 NULL characters */
	    if (memcmp(&buf[BLOCKSIZE-8], 
		       "\000\000\000\000\000\000\000\000", 8) != 0) {
		/* continue to next block */
		size += fwrite(buf, 1, BLOCKSIZE, fp);
	    } else {
		/* last block of the file */
		int i;
		for (i = BLOCKSIZE - 6; i >= 0; i--) {
		    if (buf[i] != 0) break;
		}
		size += fwrite(buf, 1, i+1, fp);
		size += fwrite(magics[current_type].tailstr, 1, 
			       magics[current_type].tailsize, fp);
		fclose(fp);
		fprintf(stderr, "%s %6d bytes(%4.1f blocks) salvaged.\n", 
			BACKSPACE, size, (double)size/BLOCKSIZE);
		fp = NULL;
	    }
	}	    
    }
    return files;
}
